import os,json
from datetime import datetime

def ensureFolderExists(folderPath):
	if not os.path.exists(folderPath):
		os.makedirs(folderPath)
	
def updateEnvironmentVariablesFromJsonConfig():
	with open('skynamo-config.json', "r") as read_file:
		config=json.load(read_file)
		os.environ['SKYNAMO_API_KEY']=config['SKYNAMO_API_KEY']
		os.environ['SKYNAMO_INSTANCE_NAME']=config['SKYNAMO_INSTANCE_NAME']
		os.environ['SKYNAMO_REGION']=config['SKYNAMO_REGION']
		if 'SKYNAMO_CACHE_REFRESH_INTERVAL' in config:
			os.environ['SKYNAMO_CACHE_REFRESH_INTERVAL']=str(config['SKYNAMO_CACHE_REFRESH_INTERVAL'])
		if 'SKYNAMO_GMAIL_SENDER' in config:
			os.environ['SKYNAMO_GMAIL_SENDER']=config['SKYNAMO_GMAIL_SENDER']
		if 'SKYNAMO_GMAIL_PASSWORD' in config:
			os.environ['SKYNAMO_GMAIL_PASSWORD']=config['SKYNAMO_GMAIL_PASSWORD']

def getDateTimeObjectFromSkynamoDateTimeStr(skynamoDateTimeStr:str):
	dateTimeStrWithoutMicroseconds=skynamoDateTimeStr[:19]+skynamoDateTimeStr[-6:]
	return datetime.strptime(dateTimeStrWithoutMicroseconds, "%Y-%m-%dT%H:%M:%S%z")

def getStringWithOnlyValidPythonVariableCharacters(string:str):
	allowedChars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_'
	newString=''
	for char in string:
		if char in allowedChars:
			newString+=char
		else:
			newString+='_'
	if newString[0] in '0123456789':
		newString='_'+newString
	##remove non-utf8 characters
	newString=newString.encode('ascii', 'ignore').decode('ascii')
	return newString

def getPathRelativeToSkynamoDataFolder(relativeFilePath:str):
	if relativeFilePath[0]=='/':
		relativeFilePath=relativeFilePath[1:]
	return f'skynamo_data/{relativeFilePath}'

def clearInstanceSpecificFolder(relativeFolderPath:str):
	finalFolderPath=getPathRelativeToSkynamoDataFolder(relativeFolderPath)
	if os.path.exists(finalFolderPath):
		for file in os.listdir(finalFolderPath):
			os.remove(os.path.join(finalFolderPath, file))

def writeToInstanceSpecificFile(relativeFilePath:str,stringContent:str):
	finalFilePath=getPathRelativeToSkynamoDataFolder(relativeFilePath)
	with open(finalFilePath, "w") as write_file:
		write_file.write(stringContent)


from ..shared.helpers import getStringWithOnlyValidPythonVariableCharacters
from typing import List

class CustomFieldArg:
	def __init__(self,customField:dict,formPrefix:str):
		customFieldType=customField['type']
		customFieldId=customField['id']
		customFieldName=getStringWithOnlyValidPythonVariableCharacters(customField['name'])
		customPropName=f'{formPrefix}c{customFieldId}_{customFieldName}'
		argType='None'
		if customFieldType=='Text Field':
			argType='str'
		elif customFieldType=='Number Field':
			argType='float'
		elif customFieldType=='Date Time Field':
			argType='datetime'
		elif customFieldType=='Single Value Enumeration Field':
			if customField['enumeration_values']!=[]:
				commaSeparatedOptions=_getCommaSeperatedEnums(customField['enumeration_values'])
				argType=f'Literal[{commaSeparatedOptions}]'
		elif customFieldType=='Multi Value Enumeration Field':
			if customField['enumeration_values']!=[]:
				commaSeparatedOptions=_getCommaSeperatedEnums(customField['enumeration_values'])
				argType=f'List[Literal[{commaSeparatedOptions}]]'
		elif customFieldType=='Single Value Hierarchical Enumeration Field':
			if customField['enumeration_values']!=[]:
				commaSeparatedOptions=_getCommaSeperatedEnumsForNestedEnums(customField['enumeration_values'])
				argType=f'Literal[{commaSeparatedOptions}]'
		elif customFieldType=='Multi Value Hierarchical Enumeration Field':
			if customField['enumeration_values']!=[]:
				commaSeparatedOptions=_getCommaSeperatedEnumsForNestedEnums(customField['enumeration_values'])
				argType=f'List[Literal[{commaSeparatedOptions}]]'
		elif customFieldType=='Address Field':
			argType='Address'
		elif customFieldType=='Single Value Lookup Entity Field':
			argType='int'
		elif customFieldType=='Multi Value Lookup Entity Field':
			argType='List[int]'
		self.argType:str=argType
		self.argName:str=customPropName
		self.required:bool=customField['required']

def _getFormPrefix(formDef):
	formType=formDef['type']
	formPrefix=''
	formId=formDef['id']
	if formType in ['Order','CreditRequest','Quote']:
		formPrefix=f'f{formId}_'
	return formPrefix

def _getCommaSeperatedEnums(enumerationValues:List[dict]):
	commaSeparatedOptions=''
	for enum in enumerationValues:
		option=enum["label"]
		##remove non-ascii characters
		option=option.encode('ascii', 'ignore').decode('ascii')
		commaSeparatedOptions+=f'"{option}",'
	return commaSeparatedOptions[:-1]

def _getCommaSeperatedEnumsForNestedEnums(enumValues:List[dict]):
	commaSeparatedOptions=''
	parentToChildEnumValues={}
	for enum in enumValues:
		
		if 'parent_id' not in enum:
			parentToChildEnumValues[enum['id']]={'label':enum['label'],'children':[]}
		else:
			parentToChildEnumValues[enum['parent_id']]['children'].append(enum)
	for parentEnumId in parentToChildEnumValues:
		for childEnum in parentToChildEnumValues[parentEnumId]['children']:
			parentOption=parentToChildEnumValues[parentEnumId]['label']
			childOption=childEnum['label']
			##remove non-ascii characters
			parentOption=parentOption.encode('ascii', 'ignore').decode('ascii')
			childOption=childOption.encode('ascii', 'ignore').decode('ascii')

			commaSeparatedOptions+=f'"{parentOption} - {childOption}",'
	return commaSeparatedOptions[:-1]

def getListCustomFieldArgs(formDef):
	customFields=formDef['custom_fields']
	skippedCustomFieldTypes=['Images Field','Signature Field','Sketch Field','Divider Field','Label Field']
	listOfCustomFieldArgs:List[CustomFieldArg]=[]
	formPrefix=_getFormPrefix(formDef)
	for customField in customFields:
		customFieldType=customField['type']
		if customFieldType in skippedCustomFieldTypes:
			continue
		listOfCustomFieldArgs.append(CustomFieldArg(customField,formPrefix))
	return listOfCustomFieldArgs


import os
from ..reader.sync import SyncDataTypesFromSkynamo,getSynchedDataTypeFileLocation
import ujson
from .InstanceClassContentsCls import InstanceClassContents

def deleteExistingCodeAndEnsureRequiredFoldersExist():
	## remove skynamoInstanceDataClasses subfolders and files
	if os.path.exists('skynamo_data'):
		if os.path.exists('skynamo_data/code'):
			for root, dirs, files in os.walk('skynamo_data/code', topdown=False):
				for name in files:
					os.remove(os.path.join(root, name))
				for name in dirs:
					os.rmdir(os.path.join(root, name))
		else:
			os.mkdir('skynamo_data/code')
		if not(os.path.exists('skynamo_data/cache')):
			os.mkdir('skynamo_data/cache')
	else:
		os.mkdir('skynamo_data')
		os.mkdir('skynamo_data/code')
		os.mkdir('skynamo_data/cache')

def refreshCustomFormsAndFields():
	deleteExistingCodeAndEnsureRequiredFoldersExist()
	SyncDataTypesFromSkynamo(['formdefinitions'])
	formsJson={}
	with open(getSynchedDataTypeFileLocation('formdefinitions'), "r") as read_file:
		formsJson=ujson.load(read_file)
	instanceClassContents=InstanceClassContents()
	for formDefId in formsJson['items']:
		formDef=formsJson['items'][formDefId]
		if formDef['active']==True:
			instanceClassContents.addFormDefinitionToRelevantClassContents(formDef)
	instanceClassContents.write()




import ujson
from ..models.Transaction import Transaction
from ..models.Address import Address
from ..shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from typing import List

def populateUserIdAndNameFromInteractionAndReturnFormIds(transaction:Transaction,interactionsJson:dict):
	interaction=interactionsJson['items'][str(transaction.interaction_id)]
	transaction.user_id=interaction['user_id']
	transaction.user_name=interaction['user_name']
	formIds=[]
	if 'completed_form_ids' in interaction:
		formIds=interaction['completed_form_ids']
	return formIds

def populateCustomPropsFromFormResults(transaction:Transaction,formIds:List[int],completedForms:dict):
	for id in formIds:
		formRes=completedForms['items'][str(id)]
		for customField in formRes['custom_fields']:
			customFieldId=customField['id']
			formId=formRes['form_id']
			customProp=getPropThatStartsWith(transaction,f'f{formId}_c{customFieldId}')
			setTypeCorrectedCustomFieldValue(transaction,customField,customProp)

def setTypeCorrectedCustomFieldValue(item:object,customField:dict,customProp:str):
	## check if customProp is a valid property of item
	if customProp not in item.__dict__:
		return
	if not 'value' in customField:
		return # don't set empty values
	typeCorrectedCustomFieldValue=customField['value']
	if typeCorrectedCustomFieldValue=='':
		return # don't set empty values
	expectedPropType=item.__getattribute__('_customFieldPropTypes')[customProp]
	if expectedPropType=='int':
		typeCorrectedCustomFieldValue=int(typeCorrectedCustomFieldValue)
	elif expectedPropType=='float':
		typeCorrectedCustomFieldValue=float(typeCorrectedCustomFieldValue)
	elif expectedPropType=='datetime':
		typeCorrectedCustomFieldValue=getDateTimeObjectFromSkynamoDateTimeStr(typeCorrectedCustomFieldValue)
	elif expectedPropType=='Address':
		adr=Address()
		adr.populateFromJsonValue(typeCorrectedCustomFieldValue)
		typeCorrectedCustomFieldValue=adr
	elif expectedPropType[:4]=='list':
		stringAr=typeCorrectedCustomFieldValue.split(',')
		if expectedPropType[5:8]=='int':
			typeCorrectedCustomFieldValue=[]
			for i in range(len(stringAr)):
				typeCorrectedCustomFieldValue.append(int(stringAr[i]))
		else:
			typeCorrectedCustomFieldValue=stringAr
	setattr(item,customProp,typeCorrectedCustomFieldValue)

def getListOfObjectsFromJsonFile(jsonFile:str,DataClass):
	formIdToFilterOn=0
	if '_f' in DataClass.__name__:
		formIdToFilterOn=int(DataClass.__name__.split('_f')[-1])
	jsonDict={}
	with open(jsonFile, "r") as read_file:
		jsonDict=ujson.load(read_file)
	listOfObjects=[]
	for itemId in jsonDict['items']:
		item=jsonDict['items'][itemId]
		if formIdToFilterOn!=0:
			if item['form_id']!=formIdToFilterOn:
				continue
		obj=DataClass(item)
		if 'custom_fields' in item:
			for customField in item['custom_fields']:
				customFieldId=customField['id']
				customProp=getPropThatStartsWith(obj,f'c{customFieldId}_')
				setTypeCorrectedCustomFieldValue(obj,customField,customProp)
		listOfObjects.append(obj)
	return listOfObjects

def getPropThatStartsWith(item:object,substringToStartWith):
	customProp=''
	for atr in item.__dict__.keys():
		if atr.startswith(substringToStartWith):
			customProp=atr
			break
	return customProp


from datetime import datetime
from typing import Literal,Union,List

from .WriteOperationCls import WriteOperation

def isBasicType(fieldValue):
	return isinstance(fieldValue,str) or isinstance(fieldValue,bool) or isinstance(fieldValue,int) or isinstance(fieldValue,float) or isinstance(fieldValue,list) or isinstance(fieldValue,dict) or fieldValue==None

def getJsonReadyFieldValue(fieldValue):
	if isinstance(fieldValue,datetime):
		return fieldValue.strftime('%Y-%m-%dT%H:%M:%S')
	elif isBasicType(fieldValue)==False:
		return fieldValue.getJsonReadyValue()
	elif isinstance(fieldValue,list):
		for i in range(len(fieldValue)):
			if isBasicType(fieldValue[i])==False:
				fieldValue[i]=fieldValue[i].getJsonReadyValue()
	return fieldValue

def getCustomFieldIdIfFieldIsCustomField(fieldName:str):
	## if format of "f{digit}_c{customFieldDigit}_*" or "c{customFieldDigit}_*", return customFieldDigit else return None
	fieldName=fieldName.split('_')[0]
	if len(fieldName)>1:
		fieldName=fieldName[1:]
		if fieldName.isdigit():
			return int(fieldName)
		else:
			fieldName=fieldName.split('_')[0]
			if len(fieldName)>1:
				fieldName=fieldName[1:]
				if fieldName.isdigit():
					return int(fieldName)
	return None


def addPatchedFieldToBodyIfAllowed(body:dict,fieldName:str,fieldValue,object:object):
	if fieldName in ['id','row_version','version','create_date','last_modified_time','tax']:
		raise Exception(f'Field {fieldName} cannot be patched')
	if fieldName not in object.__dict__:
		raise Exception(f'Field {fieldName} is not a valid {type(object).__name__} field')
	customFieldId=getCustomFieldIdIfFieldIsCustomField(fieldName)
	jsonReadyFieldValue=getJsonReadyFieldValue(fieldValue)
	if customFieldId!=None:
		if 'custom_fields' not in body:
			body['custom_fields']=[]
		body['custom_fields'].append({'id':customFieldId,'value':jsonReadyFieldValue})
	else:
		body[fieldName]=jsonReadyFieldValue

def getWriteOperationToUpdateObject(object:object,fieldsToPatch:List[str],httpMethod:Literal['patch','put','post']='patch'):
	body={'id':object.id}#type:ignore
	for fieldName in fieldsToPatch:
		fieldValue=object.__dict__[fieldName]
		if fieldValue==None:
			raise Exception(f'Error in field: {fieldName}, it is set to null. If you want to remove a value for customers or products, you need to do writer.addReplaceCustomer() or writer.addReplaceProdct()')
		addPatchedFieldToBodyIfAllowed(body,fieldName,object.__dict__[fieldName],object)
	import json
	print(json.dumps(body))
	return WriteOperation(type(object).__name__.lower()+'s', httpMethod, body)

def getWriteOperationToPutObject(object:object,httpMethod:Literal['patch','put','post']='put'):
	body=object.__dict__.copy()
	for fieldToNotUpdate in ['row_version','version','create_date','last_modified_time']:
		if fieldToNotUpdate in body:
			del body[fieldToNotUpdate]
	convertToSkynamoApiReadyValues(body)
	return WriteOperation(type(object).__name__.lower()+'s', httpMethod, body)

def addToCustomFieldsIfCustomField(key:str,body:dict):
	## if format of "f{digit}_c{customFieldDigit}_*" or "c{customFieldDigit}_*", add to custom fields
	customFieldId=getCustomFieldIdIfFieldIsCustomField(key)
	if customFieldId!=None:
		body['custom_fields'].append({'id':customFieldId,'value':body[key]})
		return True
	else:
		return False

def convertToSkynamoApiReadyValues(body:dict):
	keysToDelete=[]
	body['custom_fields']=[]
	for key in body:
		if body[key]==None:
			keysToDelete.append(key)
		else:
			if type(body[key])==datetime:
				body[key]=body[key].strftime('%Y-%m-%dT%H:%M:%S')
			elif isBasicType(body[key])==False:
				body[key]=body[key].getJsonReadyValue()
			elif isinstance(body[key],list):
				for i in range(len(body[key])):
					if isBasicType(body[key][i])==False:
						body[key][i]=body[key][i].getJsonReadyValue()
					elif isinstance(body[key][i],dict):
						convertToSkynamoApiReadyValues(body[key][i])
			elif isinstance(body[key],dict):
				convertToSkynamoApiReadyValues(body[key])

			isCustomField=addToCustomFieldsIfCustomField(key,body)
			if isCustomField:
				keysToDelete.append(key)
	for key in keysToDelete:
		del body[key]
	if len(body['custom_fields'])==0:
		del body['custom_fields']

def getBodyForWriteOperation(locals):
	body=locals
	del body['self']
	convertToSkynamoApiReadyValues(body)
	return body

def addWarehouseAndLabelToStockLevelUpdateIfPresent(item:dict,warehouse_id:int,label:Union[None,str]):
	if warehouse_id:
		item['warehouse_id'] = warehouse_id
	if label:
		item['label'] = label


from datetime import datetime
from ..shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr

class Price:
	def __init__(self,json:dict):
		self.price_list_id:int=json['price_list_id']
		self.price_list_name:str=json['price_list_name']
		self.product_id:int=json['product_id']
		self.product_code:str=json['product_code']
		self.product_name:str=json['product_name']
		self.price:float=json['price']
		self.order_unit_id:int=json['order_unit_id']
		self.order_unit_name:str=json['order_unit_name']
		self.last_modified_time:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['last_modified_time'])
		if 'tax_rate_id' in json:
			self.tax_rate_id:int=json['tax_rate_id']


from skynamo.writer.WriterBase import WriterBase
from skynamo.writer.writeHelpers import getWriteOperationToUpdateObject,getBodyForWriteOperation,getWriteOperationToPutObject
from skynamo.writer.WriteOperationCls import WriteOperation

from skynamo.models.Location import Location
from skynamo.models.OrderUnit import OrderUnit
from skynamo.models.LineItem import LineItem
from skynamo.models.Address import Address

from .Customer import Customer
from .Product import Product
from .Order import Order
from .Quote import Quote
from .CreditRequest import CreditRequest
from typing import Union,Literal,List,Dict
from datetime import datetime

class Writer(WriterBase):
	def __init__(self):
		super().__init__()
		self.writeOperations=[]
	## add update operations dependent on custom fields
	def addCustomerUpdate(self,customer:Customer,fieldsToUpdate:List[str]):
		self.writeOperations.append(getWriteOperationToUpdateObject(customer,fieldsToUpdate))
	def addCustomerReplace(self,customer:Customer):
		self.writeOperations.append(getWriteOperationToPutObject(customer))
	
	def addProductUpdate(self,product:Product,fieldsToUpdate:List[str]):
		self.writeOperations.append(getWriteOperationToUpdateObject(product,fieldsToUpdate))
	def addProductReplace(self,product:Product):
		self.writeOperations.append(getWriteOperationToPutObject(product))
	## add create operations dependent on custom fields
	def addCustomerCreate(self,code:str,name:str,##|requiredCustomerCustomFields|##,
						active:bool=True,location:Union[Location,None]=None,
						price_list_id:Union[int,None]=None,price_list_name:Union[str,None]=None,assigned_users:List[int]=[],
						default_discount:float=0.0,default_warehouse_id:Union[int,None]=None,default_warehouse_name:Union[str,None]=None,
						##|optionalCustomerCustomFields|##
						):
		self.writeOperations.append(WriteOperation("customers", "post", getBodyForWriteOperation(locals())))
	
	def addProductCreateWithSingleUnit(self,code:str,name:str,##|requiredProductCustomFields|##,
										unit_name:str='Unit',active:bool=True,##|optionalProductCustomFields|##
										):
		argsDict=locals()
		del argsDict['unit_name']
		body=getBodyForWriteOperation(argsDict)
		body['order_units']=[{'name':unit_name,'multiplier':1,'active':True}]
		self.writeOperations.append(WriteOperation("products", "post", body))

	def addProductCreateWithMultipleOrderUnits(self,code:str,name:str,order_units:List[OrderUnit],##|requiredProductCustomFields|##,
									active:bool=True,##|optionalProductCustomFields|##
										):
		self.writeOperations.append(WriteOperation("products", "post", getBodyForWriteOperation(locals())))

	def addOrderCreate(self,customer_id:int,date:datetime,user_id:int,items:List[LineItem],##|requiredOrderCustomFields|##,
						warehouse_id:Union[int,None]=None,prices_include_vat:Union[bool,None]=None,discount:Union[None,float]=None,
						##|optionalOrderCustomFields|##
						):
		self.writeOperations.append(WriteOperation("orders", "post", getBodyForWriteOperation(locals())))
	def addQuoteCreate(self,customer_id:int,date:datetime,user_id:int,items:List[LineItem],##|requiredQuoteCustomFields|##,
						warehouse_id:Union[int,None]=None,prices_include_vat:Union[bool,None]=None,discount:Union[None,float]=None,
						##|optionalQuoteCustomFields|##
						):
		self.writeOperations.append(WriteOperation("quotes", "post", getBodyForWriteOperation(locals())))
	def addCreditRequestCreate(self,customer_id:int,date:datetime,user_id:int,items:List[LineItem],##|requiredCreditRequestCustomFields|##,
						warehouse_id:Union[int,None]=None,prices_include_vat:Union[bool,None]=None,discount:Union[None,float]=None,
						##|optionalCreditRequestCustomFields|##
						):
		self.writeOperations.append(WriteOperation("creditrequests", "post", getBodyForWriteOperation(locals())))




from ..shared.helpers import ensureFolderExists
from typing import Union,Literal,List,Dict
from ..shared.api import makeRequest
import ujson,math,threading

def AddPageResultToExistingItemsAndReturnTotalItems(dataType:str,existingItems:Dict,pageNr:int,pageSize:int,filters:str,flags:List[str]=[]):
	print(f'Pagenr: {pageNr}')
	params:Dict[str,Union[str,int]]={'page_number':pageNr,'page_size':pageSize}
	if filters!='':
		params['filters']=filters
	if flags!=[]:
		params['flags']=','.join(flags)
	jsonResponse=makeRequest('get',dataType,params)
	if 'message' in jsonResponse:
		if jsonResponse['message']=='Forbidden':
			raise Exception('Invalid region, instance name or api key')
	totalItems=0
	if 'page' in jsonResponse:
		totalItems=jsonResponse['page']['filtered_item_count']
	if 'data' in jsonResponse:
		data=jsonResponse['data']
		for item in data:
			id=''
			if dataType=='stocklevels':
				warehouseId=0
				if 'warehouse_id' in item:
					warehouseId=item['warehouse_id']
				id=f'{item["product_id"]},{item["order_unit_id"]},{warehouseId}'
			elif dataType=='prices':
				id=f'{item["product_id"]},{item["order_unit_id"]},{item["price_list_id"]}'
			else:
				id=item['id']
			existingItems[id]=item
	return totalItems

def addLatestRowVersionToExistingData(existingData:dict,existingItems:dict):
	latestRowVersion=0
	for item in existingItems.values():
		possibleVersionKeys=['row_version','version']
		for key in possibleVersionKeys:
			if key in item:
				rowVersion=item[key]
				if rowVersion>latestRowVersion:
					latestRowVersion=rowVersion
	existingData['lastRowVersion']=latestRowVersion

def SyncDataTypeFromSkynamoToLocalJsonFiles(dataType,fullSync=False,syncFromLastRowVersion=False,flags:List[str]=[]):
	exceptionThatOccured=None
	print(f'Starting sync for {dataType}')
	pageSize=200
	totalItems=-1
	existingData={'lastPageNr':1,'items':{},'lastRowVersion':0}
	if not(fullSync):
		try:
			with open(getSynchedDataTypeFileLocation(dataType), "r") as read_file:
				existingData=ujson.load(read_file)
		except Exception as e:
			pass
	pageNr=1
	filters=''
	if syncFromLastRowVersion:
		rowVersion=existingData['lastRowVersion']
		filters=f'[\"greater_than(version,{rowVersion})\"]'
	else:
		pageNr=existingData['lastPageNr']
	existingItems=existingData['items']
	try:
			totalItems=AddPageResultToExistingItemsAndReturnTotalItems(dataType,existingItems,pageNr,pageSize,filters,flags)
			pageNr=pageNr+1
			predictedNumberOfPages=math.ceil(totalItems/pageSize)
			while pageNr<=predictedNumberOfPages:
				pagesLeft=predictedNumberOfPages-pageNr + 1
				nrThreads=pagesLeft
				if nrThreads>20:
					nrThreads=20
				threads=[]
				for t in range(nrThreads):
					t=threading.Thread(target=AddPageResultToExistingItemsAndReturnTotalItems,args=(dataType,existingItems,pageNr+t,pageSize,filters,flags))
					t.start()
					threads.append(t)
				for thread in threads:
					thread.join()
				pageNr=pageNr+nrThreads
	except Exception as e:
		exceptionThatOccured=e

	addLatestRowVersionToExistingData(existingData,existingItems)
	existingData['lastPageNr']=pageNr-1
	with open(getSynchedDataTypeFileLocation(dataType), "w") as write_file:
		print(f'updating cache for {dataType}')
		ujson.dump(existingData, write_file)
	if exceptionThatOccured!=None:
		raise exceptionThatOccured

def SyncDataTypesFromSkynamo(dataTypes:List[Literal['prices','warehouses','completedforms','quotes','orders','creditrequests','users','stocklevels','customers','products','invoices','formdefinitions','interactions']]=['completedforms','quotes','orders','creditrequests','users','stocklevels','customers','products','invoices','interactions']):
	fullSyncDataTypes=['users','stocklevels','formdefinitions']
	versionedDataTypes=['customers','products','invoices']
	dataTypeFlags={'formdefinitions':['show_enums'],'orders':['show_nulls'],'creditrequests':['show_nulls'],'quotes':['show_nulls']}
	flags=[]
	for dataType in dataTypes:
		fullSync=False
		if dataType in fullSyncDataTypes:
			fullSync=True
		syncFromLastRowVersion=False
		if dataType in versionedDataTypes:
			syncFromLastRowVersion=True
		if dataType in dataTypeFlags:
			flags=dataTypeFlags[dataType]
		SyncDataTypeFromSkynamoToLocalJsonFiles(dataType,fullSync,syncFromLastRowVersion,flags)

def refreshJsonFilesLocallyIfOutdated(dataTypes:List[Literal['prices','warehouses','completedforms','quotes','orders','creditrequests','users','stocklevels','customers','products','invoices','formdefinitions','interactions']],forceRefresh:bool=False):
	import os
	import time
	nrSecondsToWaitBeforeRefreshing=300
	if os.environ.get('SKYNAMO_CACHE_REFRESH_INTERVAL')!=None:
		try:
			nrSecondsToWaitBeforeRefreshing=int(os.environ.get('SKYNAMO_CACHE_REFRESH_INTERVAL')) #type: ignore
		except:
			pass

	for dataType in dataTypes:
		if os.path.exists(getSynchedDataTypeFileLocation(dataType)):
			fileLastModifiedTime = os.path.getmtime(getSynchedDataTypeFileLocation(dataType))
			if forceRefresh or time.time()-fileLastModifiedTime>nrSecondsToWaitBeforeRefreshing:
				SyncDataTypesFromSkynamo([dataType])
		else:
			SyncDataTypesFromSkynamo([dataType])

def getSynchedDataTypeFileLocation(dataType:str):
	return f'skynamo_data/cache/{dataType}.json'


from typing import Literal,List
class WriteError:
	def __init__(self,dataType:str,httpMethod:Literal['post','patch','put'],itemOrId:list,error:List[str]):
		self.dataType=dataType
		self.itemOrId=itemOrId
		self.httpMethod=httpMethod
		self.error=error


from ..shared.helpers import getStringWithOnlyValidPythonVariableCharacters

def getNameOfModelClassOnWhichToAddCustomFieldFromFormDefinition(formDef):
	formType=formDef['type']
	formId=formDef['id']
	formName=formDef['name']
	instanceClassName=getStringWithOnlyValidPythonVariableCharacters(formName)+f'_f{formId}'
	if formType in ['Order','CreditRequest','Quote']:
		instanceClassName=formType
	elif formId==-3:
		instanceClassName='Product'
	elif formId==-1:
		instanceClassName='Customer'
	return instanceClassName

def isCustomForm(instanceClassName):
	if '_f' in instanceClassName:
		return True
	return False





from datetime import datetime
from skynamo.shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from typing import Union

class StockLevel:
	def __init__(self,json:dict):
		self.product_id:int=json['product_id']
		self.product_code:str=json['product_code']
		self.product_name:str=json['product_name']
		self.order_unit_id:int=json['order_unit_id']
		self.order_unit_name:str=json['order_unit_name']
		self.warehouse_id:int=0
		if 'warehouse_id' in json:
			self.warehouse_id:int=json['warehouse_id']
		self.warehouse_name:str='Null warehouse'
		if 'warehouse_name' in json:
			self.warehouse_name:str=json['warehouse_name']
		self.level:int=json['level']
		self.label:Union[None,str]=None
		if 'label' in json:
			self.label:Union[None,str]=json['label']
		self.last_modified_time:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['last_modified_time'])